/*
	Win32MultipleDragList.c

	Copyright  2000, 2001 Kyle Hammond

	All rights reserved.

	March 21, 2000	Initial release.
					Search for 'stub' to find places to change the cursor.  Since Win32 doesn't have default cursors
						for this stuff, you'll have to come up with your own.

	Best viewed in Geneva 9 point.
*/
#if !defined( _INC_COMMCTRL )
	#include <CommCtrl.h>
#endif
#if !defined( _INC_WINDOWSX )
	#include <WindowsX.h>
#endif
#include "Win32MultipleDragList.h"

#define kMyDragProperty			"PROP_MULTIPLE_DRAG_LIST"
#define kMyDragCursorProperty	"PROP_MY_CURSOR_DRAGGING"
#define IDT_MyTimer			66

static LRESULT CALLBACK local_windowProc( IN HWND inHWnd, IN UINT inMsg, IN WPARAM inWParam, IN LPARAM inLParam );
static void local_sendCancelDragNotification( HWND inListHWnd );
static void local_sendDraggingNotification( HWND inListHWnd, POINT *inPoint );

static UINT		_gDragListMsg = 0;

BOOL MakeMultipleDragList( HWND inListHWnd )
{	HGLOBAL		tempHandle;
	LPVOID		tempPtr;

	/* Determine what the DRAGLISTMSGSTRING message value is. */
	if ( _gDragListMsg == 0 )
		_gDragListMsg = RegisterWindowMessage( DRAGLISTMSGSTRING );

	/* Now attach a property that contains the original WNDPROC of the list. */
	tempHandle = GlobalAlloc( GPTR, sizeof(WNDPROC) );
	if ( tempHandle != NULL ) {
		tempPtr = GlobalLock( tempHandle );
		if ( tempPtr != NULL ) {
			ZeroMemory( tempPtr, sizeof(WNDPROC) );
			*(WNDPROC *)tempPtr = (WNDPROC)SetWindowLong( inListHWnd, GWL_WNDPROC, (LONG)local_windowProc );
			GlobalUnlock( tempHandle );

			if ( SetProp( inListHWnd, kMyDragProperty, tempHandle ) != 0 )
				return true;
		}
	}

	/* If anything above failed, we return false to signify that we couldn't make the list a MultipleDragList. */
	return false;
}

LRESULT CALLBACK local_windowProc( IN HWND inHWnd, IN UINT inMsg, IN WPARAM inWParam, IN LPARAM inLParam )
{	LRESULT		result = 1;
	LONG			tempLong, tempLong2;
	POINT		tempPoint;
	DRAGLISTINFO	dragInfo;
	HGLOBAL		tempHandle;
	LPVOID		tempPtr;

	switch ( inMsg ) {
	case WM_LBUTTONDOWN:
		/* First, determine how many cells are selected. */
		tempLong = SendMessage( inHWnd, LB_GETSELCOUNT, 0, 0 );
		if ( tempLong == LB_ERR ) {
			/* The list is a single-selection list box. */
			tempLong = SendMessage( inHWnd, LB_GETCURSEL, 0, 0 );
			if ( tempLong == LB_ERR )
				tempLong = 0;
			else if ( tempLong >= 0 )
				tempLong = 1;
		}

		if ( tempLong > 0 ) {
			/* At least one cell is selected in the list. */

			/* Convert the point to screen coordinates for LBItemFromPt. */
			tempPoint.x = GET_X_LPARAM( inLParam );
			tempPoint.y = GET_Y_LPARAM( inLParam );
			MapWindowPoints( inHWnd, NULL, &tempPoint, 1 );

			tempLong2 = LBItemFromPt( inHWnd, tempPoint, false );

			/* Determine if the user clicked in a selected cell. */
			if ( SendMessage( inHWnd, LB_GETSEL, tempLong2, 0 ) > 0 ) {
				/* Start a drag here. */
				dragInfo.uNotification = DL_BEGINDRAG;
				dragInfo.hWnd = inHWnd;
				dragInfo.ptCursor = tempPoint;

				if ( SendMessage( (HWND)GetWindowLongPtr( inHWnd, GWLP_HWNDPARENT ), _gDragListMsg,
							GetDlgCtrlID( inHWnd ), (LPARAM)&dragInfo ) ) {
					/* If parent window says OK, detect if the user started dragging. */

					if ( DragDetect( inHWnd, tempPoint ) ) {
						/* Put a property in the window that shows we are dragging. */
						tempHandle = GlobalAlloc( GPTR, sizeof(HCURSOR) );
						if ( tempHandle != NULL ) {
							tempPtr = GlobalLock( tempHandle );
							ZeroMemory( tempPtr, sizeof(HCURSOR) );
							GlobalUnlock( tempHandle );

							if ( SetProp( inHWnd, kMyDragCursorProperty, tempHandle ) != 0 ) {
								/* Start a timer going; 20 pulses per second. */
								SetTimer( inHWnd, IDT_MyTimer, 50, NULL );

								result = 0;
							} /* else SetProp failed */
						} /* else couldn't allocate tempHandle */
					} /* else if user doesn't move mouse, we're not dragging. */
				} /* else if parent window vetos the drag, we're not dragging. */
			} /* else if we're not clicking in a selected cell, we're not dragging. */
		} /* else no cells are selected; we're not dragging. */
		break;

	case WM_RBUTTONDOWN:
		/* Right button down during a drag cancels the drag. */
		if ( GetProp( inHWnd, kMyDragProperty ) != NULL ) {
			local_sendCancelDragNotification( inHWnd );
			result = 0;
		}
		break;

	case WM_CHAR:
		/* Esc key down during a drag cancels the drag. */
		if ( inWParam == VK_ESCAPE && GetProp( inHWnd, kMyDragProperty ) != NULL ) {
			local_sendCancelDragNotification( inHWnd );
			result = 0;
		}
		break;

	case WM_TIMER:
		if ( inWParam == IDT_MyTimer && GetProp( inHWnd, kMyDragCursorProperty ) != NULL ) {
			/* We are dragging items around. */
			DWORD	tempWord = GetMessagePos( );

			if ( !GetCursorPos( &tempPoint ) ) {
				tempPoint.x = GET_X_LPARAM( tempWord );
				tempPoint.y = GET_Y_LPARAM( tempWord );
			}

			local_sendDraggingNotification( inHWnd, &tempPoint );

			result = 0;
		}
		break;

	case WM_MOUSEMOVE:
		tempPoint.x = GET_X_LPARAM( inLParam );
		tempPoint.y = GET_Y_LPARAM( inLParam );
		MapWindowPoints( inHWnd, NULL, &tempPoint, 1 );

		if ( GetProp( inHWnd, kMyDragCursorProperty ) != NULL ) {
			/* We are dragging items around. */
			local_sendDraggingNotification( inHWnd, &tempPoint );

			result = 0;
		} else {
			/* We're not dragging but we may be floating above a selection that we can drag. */
			tempLong2 = LBItemFromPt( inHWnd, tempPoint, false );

			/* Determine if we're floating above a selected cell. */
			if ( SendMessage( inHWnd, LB_GETSEL, tempLong2, 0 ) > 0 )
/* stub - make cursor an open hand.
				SetCursor( LoadCursor( (HINSTANCE)GetWindowLongPtr( inHWnd, GWLP_HINSTANCE ), "OpenHand" ) );
*/				;
			else
				SetCursor( LoadCursor( NULL, IDC_ARROW ) );
		}
		break;

	case WM_LBUTTONUP:
		tempHandle = GetProp( inHWnd, kMyDragCursorProperty );
		if ( tempHandle != NULL ) {
			/* We were dragging items around. */
			KillTimer( inHWnd, IDT_MyTimer );

			tempPtr = GlobalLock( tempHandle );
			if ( tempPtr != NULL )
				SetCursor( *(HCURSOR *)tempPtr );
			GlobalUnlock( tempHandle );

 			GlobalFree( RemoveProp( inHWnd, kMyDragCursorProperty ) );

			tempPoint.x = GET_X_LPARAM( inLParam );
			tempPoint.y = GET_Y_LPARAM( inLParam );
			MapWindowPoints( inHWnd, NULL, &tempPoint, 1 );

			/* This mimics Win32 SDK behavior.
				Personally, I'd determine if the button up occurred while the list items were within the list.
				If the drop was in the list then return DL_DROPPED, otherwise use DL_CANCELDRAG.
				But, that's personal preference. */
			dragInfo.uNotification = DL_DROPPED;
			dragInfo.hWnd = inHWnd;
			dragInfo.ptCursor = tempPoint;

 			SendMessage( (HWND)GetWindowLongPtr( inHWnd, GWLP_HWNDPARENT ), _gDragListMsg,
 						GetDlgCtrlID( inHWnd ), (LPARAM)&dragInfo );

			result = 0;
		}
		break;

	case WM_DESTROY:
		/* Remove the dragging property if it's still there.  It shouldn't be but just be sure. */
		if ( GetProp( inHWnd, kMyDragCursorProperty ) != NULL )
			local_sendCancelDragNotification( inHWnd );

		// This list is no longer a MultipleDragList.
		GlobalFree( RemoveProp( inHWnd, kMyDragProperty ) );
		break;
	}

	if ( result == 1 ) {
		/* Call the original window proc if we didn't handle it here. */
		tempHandle = GetProp( inHWnd, kMyDragProperty );
		if ( tempHandle != NULL ) {
			WNDPROC		tempWndProc = NULL;

			tempPtr = GlobalLock( tempHandle );
			if ( tempPtr != NULL )
				tempWndProc = *(WNDPROC *)tempPtr;
			GlobalUnlock( tempHandle );

			if ( tempWndProc != NULL )
				result = CallWindowProc( tempWndProc, inHWnd, inMsg, inWParam, inLParam );
		}
	}

	return result;
}

static void local_sendCancelDragNotification( HWND inListHWnd )
{	HGLOBAL		tempHandle;
	LPVOID		tempPtr;
	DWORD		tempWord = GetMessagePos( );
	DRAGLISTINFO	dragInfo;

	KillTimer( inListHWnd, IDT_MyTimer );

	tempHandle = GetProp( inListHWnd, kMyDragCursorProperty );
	if ( tempHandle != NULL ) {
		tempPtr = GlobalLock( tempHandle );
		if ( tempPtr != NULL )
			SetCursor( *(HCURSOR *)tempPtr );
		GlobalUnlock( tempHandle );
	}

	GlobalFree( RemoveProp( inListHWnd, kMyDragCursorProperty ) );

	dragInfo.uNotification = DL_CANCELDRAG;
	dragInfo.hWnd = inListHWnd;
	if ( !GetCursorPos( &dragInfo.ptCursor ) ) {
		dragInfo.ptCursor.x = GET_X_LPARAM( tempWord );
		dragInfo.ptCursor.y = GET_Y_LPARAM( tempWord );
	}

	SendMessage( (HWND)GetWindowLongPtr( inListHWnd, GWLP_HWNDPARENT ), _gDragListMsg, GetDlgCtrlID( inListHWnd ),
				(LPARAM)&dragInfo );
}

static void local_sendDraggingNotification( HWND inListHWnd, POINT *inPoint )
{	DRAGLISTINFO	dragInfo;
	HCURSOR		prevCursor = NULL;
	HGLOBAL		dataH;
	LPVOID		originalCursor;

	dragInfo.uNotification = DL_DRAGGING;
	dragInfo.hWnd = inListHWnd;
	dragInfo.ptCursor = *inPoint;

	switch ( SendMessage( (HWND)GetWindowLongPtr( inListHWnd, GWLP_HWNDPARENT ), _gDragListMsg,
				GetDlgCtrlID( inListHWnd ), (LPARAM)&dragInfo ) ) {
	/* Change cursor depending on parent return value. */
	case DL_STOPCURSOR:
		prevCursor = SetCursor( LoadCursor( NULL, IDC_NO ) );
		break;
	case DL_COPYCURSOR:
/* stub - fill in your own ClosedHand with a plus cursor here.
		SetCursor( LoadCursor( (HINSTANCE)GetWindowLongPtr( inHWnd, GWLP_HINSTANCE ), "ClosedHandWithPlus" ) );
*/		prevCursor = SetCursor( LoadCursor( NULL, IDC_ARROW ) );
		break;
	case DL_MOVECURSOR:
/* stub - fill in your own ClosedHand cursor here.
		SetCursor( LoadCursor( (HINSTANCE)GetWindowLongPtr( inHWnd, GWLP_HINSTANCE ), "ClosedHand" ) );
*/		prevCursor = SetCursor( LoadCursor( NULL, IDC_ARROW ) );
		break;
	}

	/* Store the original cursor in the drag property for restoration later. */
	dataH = GetProp( inListHWnd, kMyDragCursorProperty );
	if ( dataH != NULL ) {
		originalCursor = GlobalLock( dataH );
		if ( originalCursor == NULL && prevCursor != NULL )
			*(HCURSOR *)originalCursor = prevCursor;
		GlobalUnlock( dataH );
	}
}
